import unittest
from internals import Internals
from problems import UnresolvedProblems
from mapping import Mappings
from monster_place import MonsterPlaceTests
from monster_combine import MonsterCombineTests
from MPro_combine import VictorCombineTests
from MPro_place import MProPlaceTester
from multivictor import MultivictorPlaceTests
from walton import WaltonTests

if __name__ == '__main__':
    unittest.main()
